;************************************************************************
;*                   Quicksort Benchmark VB vs. Assembler               *
;*                  by Daniel Aue for activevb.de 3.9.2003              *
;*                                                                      *
;*                     USE ENTIRELY AT YOUR OWN RISK.                   *
;*                 you may publish this source as long as               *
;*                      this commentblock stays in                      *
;************************************************************************

;************************************************************************
; Revised by  (softKUS) - III/2004
;
;             - var types vbDate, vbCurrency, vbBoolean added
;             - descending order available for all types (flags, bit 0)
;             - sort table not longer compiled, but always
;               given as parameter
;             - ref-array initialization changed in order to
;               reflect LBounds correctly
;
;        chr  - array/string may now include vbNullStrings/vbNullChars
;
;        int  - unsigned integer support added (flags, bit 1)
;               (vbBoolean, vbByte, vbInteger, vbLong, vbCurrency)
;
;        flt  - no use of FPU
;
;
;        caution:    - start/end parameters not supported anymore 
;                      arrSRT always sorts the whole array
;                    - string sort about 10 % slower than before due to
;                      interpreting BSTR headers
;
;        ErrorCodes: -6: not an array
;                    -5: no dimension (no savearray-structure available)
;                    -4: invalid dimension (<> 1)
;                    -3: no data pointer (shouldn't happen)
;                    -2: unsupported varType
;                    -1: invalid dimension (UBound = -1)
;
;************************************************************************

386
MODEL FLAT

CODE
_run:

include         ..\stdparas.inc                         ; PA1-PA4

; *********************************************************************************************

arrptr          EQU     <PA1>                           ; ptr: array to be sorted
refarr          EQU     <PA2>                           ; ptr: reference (element numbers)
srttbl          EQU     <PA3>                           ; ptr: sort table (chr sort only)
flags           EQU     <PA4>                           ; lng: Bit 0: 0=ascending
                                                        ;             1=descending
                                                        ;      Bit 1: 0=signed
                                                        ;             1=unsigned

; *********************************************************************************************

offHere         EQU     <stack1>                        ; absolute address of "here:"
errCD           EQU     <stack2>                        ; error code
srtfnc          EQU     <stack3>                        ; acc. sort function address
arrcnt          EQU     <stack4>                        ; number of elements (-1)
tmpfigH         EQU     <stack5>                        ; temp values used by
tmpfigL         EQU     <stack6>                        ; dbl/flt/chr routines

maxVarType      EQU     17                              ; number of supported types
actLB           EQU     <DWord Ptr [esp+12]>            ; ret/ret/actUB/actLB
actUB           EQU     <DWord Ptr [esp+8]>

vbArray         EQU     2000h                           ; VB types
vbPointer       EQU     4000h

var_Type        EQU     0                               ; VARIANT-structure
var_DatOff      EQU     8

sar_cDims       EQU     0                               ; SAVEARRAY-structure
sar_pvData      EQU     12
sar_Dim1UBound  EQU     16
sar_Dim1LBound  EQU     20

jmpL            EQU     7Ch                             ; jmp-commands
jmpG            EQU     7Fh                             ; set acc. to flags: signed + descending
jmpB            EQU     72h
jmpA            EQU     77h

eqvL            EQU     0
eqvG            EQU     1
eqvB            EQU     2
eqvA            EQU     3

; *********************************************************************************************

arrSRT          PROC    NEAR
                mac     SAVE
                call    here                            ; => offHere
here:           push    -6                              ; => errCD
                
chkARR:         mov     esi,arrptr                      ; "get array info"
                movzx   eax,Word Ptr [esi+var_Type]
                test    ah,vbArray/256
                jz      finish                          ; -> not an array
                inc     errCD

                mov     esi,[esi+var_DatOff]
                test    ah,vbPointer/256
                jz      chkARR_1
                mov     esi,[esi]
chkARR_1:       or      esi,esi
                jz      finish                          ; -> no dimension (no sar-structure)
                inc     errCD

                cmp     Word Ptr [esi+sar_cDims],1
                jnz     finish                          ; -> invalid dimension
                inc     errCD
                
                mov     edi,[esi+sar_pvData]
                or      edi,edi
                jz      finish                          ; -> no data pointer
                inc     errCD
                
chkTYP:         cmp     al,maxVarType +1                ; "check type/get srtfnc-address"
                jnc     finish                          ; -> varType not supported
                mov     ebx,OffSet dbTypes-OffSet here
                add     ebx,offHere
                xlat
                or      al,al
                jz      finish                          ; -> varType not supported                
                inc     errCD

                lea     edx,[ebx+(OffSet dbJumps - OffSet dbTypes)]
                cmp     al,5
                jnc     chkTYP1                         ; flag/Bit 1: not defined for dbl, ...
                test    flags,2
                jz      chkTYP1
                add     edx,(OffSet dbJumps_ns - OffSet dbJumps)

chkTYP1:        xor     ah,ah                
                mov     ax,[ebx+2*eax+(-2)+maxVarType+1]
                add     ebx,eax
                push    ebx                             ; => srtfnc

chkDIM:         mov     ecx,[esi+sar_Dim1UBound]
                dec     ecx
                js      finish                          ; -> (-1)-Dimension
                inc     errCD
                jecxz   finish                          ; -> nothing to do (only one element)
                push    ecx                             ; => arrcnt (UBound)

chkIDX:         mov     eax,[esi+sar_Dim1LBound]        ; "initialize ref-array"
                add     eax,ecx
                mov     esi,refarr
                or      esi,esi
                jmp     chkIDX_2
                            
chkIDX_1:       mov     [esi+4*ecx],eax
                dec     eax
                dec     ecx
chkIDX_2:       jnz     chkIDX_1
                
                
chkASC:         mov     esi,ebx                         ; "set jump commands acc. to flags"
                mov     ecx,flags
                and     cl,1
                xor     eax,eax
chkASC1:        dec     ebx
                mov     al,[ebx]
                or      eax,eax
                jz      doit
                add     esi,eax
                dec     ebx
                dec     ebx
                mov     al,cl
                xlat
                mov     al,[edx+eax]
                mov     [esi],al
                jmp     chkASC1
                                
doit:           xor     eax,eax                         ; "start sort process"
                mov     edx,arrcnt
                push    eax                             ; => tmpfigH
                push    edx                             ; => tmpfigL
                call    jmpSort
finish:         mov     eax,errCD
                mac     RESTSPAX


dbTypes         db      00,00,02,03                     ; nn, vbNull, vbInteger, vbLong
                db      05,06,04,06                     ; vbSingle, vbDouble, vbCurrency, vbDate
                db      07,00,00,03                     ; vbString, vbObject, vbError, vbBoolean
                db      00,00,00,00                     ; vbVariant, vbDataObject, vbDecimal, nn
                db      00,01                           ; nn, vbByte

                dw      OffSet srtLN1 - OffSet dbTypes  ; 1: one   byte sort
                dw      OffSet srtLN2 - OffSet dbTypes  ; 2: two   byte sort
                dw      OffSet srtLN4 - OffSet dbTypes  ; 3: four  byte sort
                dw      OffSet srtLN8 - OffSet dbTypes  ; 4: eight byte sort
                dw      OffSet srtSGL - OffSet dbTypes  ; 5: four  byte sort (real 4)
                dw      OffSet srtDBL - OffSet dbTypes  ; 6: eight byte sort (real 8)
                dw      OffSet srtCHR - OffSet dbTypes  ; 7: two   byte sort (1 significant)

dbJumps         db      jmpL,jmpG,jmpB,jmpA             ; "signed"   jumps
dbJumps_ns      db      jmpB,jmpA,jmpB,jmpA             ; "unsigned" jumps



; -------------------------------------------------------------------------------------
; * BYTE                
; -------------------------------------------------------------------------------------

                db      00
                db      eqvL, eqvG, OffSet srtLN1_j2 - OffSet srtLN1_j1
                db      eqvG, eqvL, OffSet srtLN1_j1 - OffSet srtLN1

srtLN1:         mov     bl,[edi+ecx]
                dec     eax
srtLN1_a:       inc     eax                
srtLN1_b:       cmp     bl,[edi+eax]
srtLN1_j1:      jg      srtLN1_a
                inc     edx                
srtLN1_c:       dec     edx
                cmp     bl,[edi+edx]
srtLN1_j2:      jl      srtLN1_c

                cmp     eax,edx
                jge     srtLN1_d
                mov     cl,[edi+eax]
                xchg    cl,[edi+edx]
                mov     [edi+eax],cl
srtLN1_d:       call    rouLoop
                jmp     srtLN1_b



; -------------------------------------------------------------------------------------
; * WORD
; -------------------------------------------------------------------------------------

                db      00
                db      eqvL, eqvG, OffSet srtLN2_j2 - OffSet srtLN2_j1
                db      eqvG, eqvL, OffSet srtLN2_j1 - OffSet srtLN2

srtLN2:         mov     bx,[edi+2*ecx]
                dec     eax
srtLN2_a:       inc     eax                
srtLN2_b:       cmp     bx,[edi+2*eax]
srtLN2_j1:      jg      srtLN2_a
                inc     edx                
srtLN2_c:       dec     edx
                cmp     bx,[edi+2*edx]
srtLN2_j2:      jl      srtLN2_c

                cmp     eax,edx
                jnc     srtLN2_d
                mov     cx,[edi+2*eax]
                xchg    cx,[edi+2*edx]
                mov     [edi+2*eax],cx
srtLN2_d:       call    rouLoop
                jmp     srtLN2_b



; -------------------------------------------------------------------------------------
; * DWORD
; -------------------------------------------------------------------------------------

                db      00
                db      eqvL, eqvG, OffSet srtLN4_j2 - OffSet srtLN4_j1
                db      eqvG, eqvL, OffSet srtLN4_j1 - OffSet srtLN4

srtLN4:         mov     ebx,[edi+4*ecx]
                dec     eax
srtLN4_a:       inc     eax                
srtLN4_b:       cmp     ebx,[edi+4*eax]
srtLN4_j1:      jg      srtLN4_a
                inc     edx                
srtLN4_c:       dec     edx
                cmp     ebx,[edi+4*edx]
srtLN4_j2:      jl      srtLN4_c

                call    rouLoop4
                jmp     srtLN4_b



; -------------------------------------------------------------------------------------
; * QWORD
; -------------------------------------------------------------------------------------

                db      00
                db      eqvB, eqvA, OffSet srtLN8_j4 - OffSet srtLN8_j3
                db      eqvL, eqvG, OffSet srtLN8_j3 - OffSet srtLN8_j2
                db      eqvA, eqvB, OffSet srtLN8_j2 - OffSet srtLN8_j1
                db      eqvG, eqvL, OffSet srtLN8_j1 - OffSet srtLN8

srtLN8:         mov     esi,[edi+8*ecx+4]
                mov     ebx,[edi+8*ecx]
                dec     eax
srtLN8_a:       inc     eax                
srtLN8_b:       cmp     esi,[edi+8*eax+4]
srtLN8_j1:      jg      srtLN8_a
                jne     srtLN8_c1
                cmp     ebx,[edi+8*eax]
srtLN8_j2:      ja      srtLN8_a
                inc     edx
srtLN8_c:       dec     edx
srtLN8_c1:      cmp     esi,[edi+8*edx+4]
srtLN8_j3:      jl      srtLN8_c
                jne     srtLN8_c2
                cmp     ebx,[edi+8*edx]
srtLN8_j4:      jb      srtLN8_c
                
srtLN8_c2:      call    rouLoop8
                jmp     srtLN8_b



; -------------------------------------------------------------------------------------
; * REAL4
; -------------------------------------------------------------------------------------

                db      00
                db      eqvL, eqvG, OffSet srtSGL_j2 - OffSet srtSGL_J1
                db      eqvG, eqvL, OffSet srtSGL_j1 - OffSet srtSGL

srtSGL:         mov     ebx,[edi+4*ecx]
                dec     eax
srtSGL_a1:      inc     eax          
srtSGL_a2:      mov     esi,ebx
                mov     ecx,[edi+4*eax]
                test    esi,ecx
                jns     srtSGL_a3
                xchg    esi,ecx
srtSGL_a3:      cmp     esi,ecx
srtSGL_j1:      jg      srtSGL_a1
                inc     edx
srtSGL_b1:      dec     edx
                mov     esi,ebx
                mov     ecx,[edi+4*edx]
                test    esi,ecx
                jns     srtSGL_b2
                xchg    esi,ecx
srtSGL_b2:      cmp     esi,ecx
srtSGL_j2:      jl      srtSGL_b1

                call    rouLoop4
                jmp     srtSGL_a2



; -------------------------------------------------------------------------------------
; * REAL8
; -------------------------------------------------------------------------------------

                db      00
                db      eqvB, eqvA, OffSet srtDBL_j4 - OffSet srtDBL_j3
                db      eqvL, eqvG, OffSet srtDBL_j3 - OffSet srtDBL_j2
                db      eqvA, eqvB, OffSet srtDBL_j2 - OffSet srtDBL_j1
                db      eqvG, eqvL, OffSet srtDBL_j1 - OffSet srtDBL

srtDBL:         mov     ebx,[edi+8*ecx]
                mov     esi,[edi+8*ecx+4]
                mov     tmpfigL,ebx
                mov     tmpfigH,esi
                dec     eax
srtDBL_a1:      inc     eax
srtDBL_a2:      lea     esi,tmpfigL
                lea     ebx,[edi+8*eax]
                mov     ecx,[esi+4]
                test    ecx,[ebx+4]
                jns     srtDBL_a3
                xchg    esi,ebx
srtDBL_a3:      mov     ecx,[esi+4]
                cmp     ecx,[ebx+4]
srtDBL_j1:      jg      srtDBL_a1
                jne     srtDBL_b2
                mov     ecx,[esi]
                cmp     ecx,[ebx]                
srtDBL_j2:      ja      srtDBL_a1
                inc     edx
srtDBL_b1:      dec     edx
srtDBL_b2:      lea     esi,tmpfigL
                lea     ebx,[edi+8*edx]
                mov     ecx,[ebx+4]
                test    ecx,[esi+4]
                jns     srtDBL_b3
                xchg    esi,ebx
srtDBL_b3:      mov     ecx,[esi+4]
                cmp     ecx,[ebx+4]
srtDBL_j3:      jl      srtDBL_b1
                jne     srtDBL_c
                mov     ecx,[esi]
                cmp     ecx,[ebx]
srtDBL_j4:      jb      srtDBL_b1                
                
srtDBL_c:       call    rouLoop8
                jmp     srtDBL_a2



comment #
; -------------------------------------------------------------------------------------
; * REAL8       Using FPU is a bit slower on my system (AMD 800) !?
; -------------------------------------------------------------------------------------

                db      00
                db      eqvB, eqvA, OffSet srtFLT_j2 - OffSet srtFLT_j1
                db      eqvA, eqvB, OffSet srtFLT_j1 - OffSet srtFLT

srtFLT:         mov     esi,[edi+8*ecx+4]
                mov     ebx,[edi+8*ecx]
                mov     tmpfigH,esi
                mov     tmpfigL,ebx
srtFLT_a:       fld     tmpfig
                mov     ebx,eax
                dec     ebx
srtFLT_b:       inc     ebx
srtFLT_c:       fcom    QWord Ptr [edi+8*ebx]
                fstsw   ax
                fwait
                sahf
srtFLT_j1:      ja      srtFLT_b
                inc     edx
srtFLT_d:       dec     edx     
                fcom    QWord Ptr [edi+8*edx]
                fstsw   ax
                fwait
                sahf
srtFLT_j2:      jb      srtFLT_d

                ffree   st(0)
                mov     eax,ebx
                call    rouLoop8
                jmp     srtFLT_a
#



; -------------------------------------------------------------------------------------
; * CHR
; -------------------------------------------------------------------------------------

                db      00
                db      eqvB, eqvA, OffSet srtCHR_j2 - OffSet srtCHR_j1
                db      eqvA, eqvB, OffSet srtCHR_j1 - OffSet srtCHR

srtCHR:         mov     ebx,srttbl
                mov     esi,[edi+4*ecx]
                mov     tmpfigH,esi
                or      esi,esi
                jz      srtCHR_x
                mov     esi,[esi-4]
srtCHR_x:       mov     tmpfigL,esi                
                dec     eax
srtCHR_a:       inc     eax                
srtCHR_b:       mov     ecx,eax
                call    chrROU
srtCHR_j1:      ja      srtCHR_a
                inc     edx                
srtCHR_c:       dec     edx
                mov     ecx,edx
                call    chrROU
srtCHR_j2:      jb      srtCHR_c
                call    rouLoop4
                jmp     srtCHR_b


chrROU:         push    edi
                push    edx
                push    eax
                mov     esi,tmpfigH
                mov     edi,[edi+4*ecx]
                mov     ecx,edi
                jecxz   chrROU_a
                mov     ecx,[edi-4]
chrROU_a:       cmp     tmpfigL,ecx
                pushfd
                jnc     chrROU_b
                mov     ecx,tmpfigL
chrROU_b:       shr     ecx,1
                jz      chrROU_e
                xor     eax,eax
                mov     edx,eax
chrROU_c:       mov     al,[esi]
                mov     dl,[edi]
                mov     al,[ebx+eax]
                inc     esi
                inc     esi
                inc     edi
                inc     edi
                cmp     al,[ebx+edx]
                jnz     chrROU_f
                dec     ecx
                jnz     chrROU_c
chrROU_e:       popfd
                push    eax
chrROU_f:       pop     eax
                pop     eax
                pop     edx
                pop     edi
                ret   

rouLoop8:       cmp     eax,edx
                jge     rouLoop
                mov     ecx,[edi+8*eax+4]
                xchg    ecx,[edi+8*edx+4]
                mov     [edi+8*eax+4],ecx
                mov     ecx,[edi+8*eax]
                xchg    ecx,[edi+8*edx]
                mov     [edi+8*eax],ecx
                jmp     rouLoop

rouLoop4:       cmp     eax,edx
                jge     rouLoop
                mov     ecx,[edi+4*eax]
                xchg    ecx,[edi+4*edx]
                mov     [edi+4*eax],ecx
                                
rouLoop:        jz      rouLoop_a
                jg      rouLoop_b
                cmp     refarr,0
                jz      rouLoop_a
                push    ebx
                mov     ebx,refarr
                mov     ecx,[ebx+4*eax]
                xchg    ecx,[ebx+4*edx]
                mov     [ebx+4*eax],ecx
                pop     ebx
rouLoop_a:      inc     eax
                dec     edx
                cmp     eax,edx
                jg      rouLoop_b
                ret
                
rouLoop_b:      cmp     edx,actLB
                jle     rouLoop_c
                mov     eax,actLB
                call    jmpSort
rouLoop_c:      cmp     eax,actUB
                jge     rouLoop_d
                mov     edx,actUB
                call    jmpSort
rouLoop_d:      pop     ecx
                ret     8

jmpSort:        pop     ecx
                push    eax
                push    edx
                push    ecx
                push    srtfnc
                lea     ecx,[eax+edx]
                shr     ecx,1
                ret
arrSRT          ENDP                
end _run
